/*
 * forwardMain.c
 *
 *  Created on: May 16, 2011
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "utilities.h"
#include "errorHandler.h"
#include "coramfVersion.h"
#include "parserVersion.h"
#include "inversionVersion.h"
#include "Inversion.h"
#include "ProblemDimension.h"
#include "ForwardErrorMessages.h"

/**
 * \page Forward Forward Calculation
 * <B>Forward Calculations </B>
 *
 * @see <a href="../ForwardNotes.pdf">Equations and specifics</a>
 * @see forwardMain.c
 */

/**
 * Check dimensions of matrices before performing calculations
 * @param[in] K kernel matrix
 * @param[in] b model matrix
 * @param[in] A noise amplitude matrix
 * @param[in] Theta dz matrix
 * @param[in] useFourierTransform True if matrices are in frequency domain.
 * @return True if dimensions are ok, false otherwise.
 */
bool DimensionsOkFwd(MAP_DEPTH_MATRIX K, MAP_DEPTH_MATRIX b, MAP_DEPTH_MATRIX A, MATRIX Theta, bool useFourierTransform) {
	char cmsg[MAX_ERROR_MSG_LENGTH];
	bool allOk = true;
	int N, Ny, Nz, Nalpha, M;
	// pull dimensions from K, then compare other matrices to K.
	N = K.nX;
	Nz = K.nZ;
	M = K.nT;
	Nalpha = K.nV;

	if (useFourierTransform) {
		Ny = K.nX; // because nX and nY were checked to be equal during fits read; don't use this: (K.nY-1) * 2;
	} else
		Ny = K.nY;

	// check that xy-plane is square
	if (N != Ny) {
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, XYPlaneMustBeSquareKernel, N, Ny, Nz, M, Nalpha);
		reportError(cmsg, "forwardMain.DimensionOK", FATAL);
		allOk = false;
	}
	if ((b.nX != N) && ((b.nY-1)*2 != N)) {
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, XYPlaneMustBeSquareModel, N, Ny, Nz, Nalpha);
		reportError(cmsg, "forwardMain.DimensionOK", FATAL);
		allOk = false;
	}

	// check for Amplitude Nx, Ny agree with Kernel, Nx, Ny
	if ((A.nX != N) || (A.nY != N) || (A.nZ != 1) || (A.nT != M)) {
		printf("A dim is (%d x %d x %d x %d\n", A.nX, A.nY, A.nZ, A.nT );
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, XYPlaneMustBeSquareNoiseAmp, N, Ny, 1, M);
		reportError(cmsg, "forwardMain.DimensionOK", FATAL);
		allOk = false;
	}

	// b should be (Nz*Nalpha)
	if (b.nZ != Nz) {
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, ModelKernelSameZ, b.nZ, Nz);
		reportError(cmsg, "forwardMain.DimensionOK", FATAL);
		allOk = false;
	}
	if (b.nV != Nalpha) {
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, ModelKernelSameAlpha, b.nV, Nalpha);
		reportError(cmsg, "forwardMain.DimensionOK", FATAL);
		allOk = false;
	}

	// Theta should be (Nz*Nalpha) x (Nz*Nalpha)
	if ((Theta.nRow != Theta.nCol) || (Theta.nCol != (Nz*Nalpha))) {
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, ExpectedThetaMatrixDim, Nz, Nalpha, Nz*Nalpha);
		reportError(cmsg, "forwardMain.DimensionOK", FATAL);
		allOk = false;
	}
	return allOk;

}

/**
 * Perform forward calculation using the input matrices and intermediate
 * matrices.  NOTE: Kernels must already have overlap matrix applied.
 * Put the noise free maps (dTilda) in intermediate values, so they
 * may be reused when adding noise.
 * @param[in,out] inversionData Pointer to structure containing all inversion data.
 */
void ForwardCalc(InversionData *inversionData) {
	InputData *input;
	input = inversionData->InputMatrices;
	TempData *tempd;
	tempd = inversionData->IntermediateMatrices;

	MAP_DEPTH_MATRIX dTildeWave;
	MAP_DEPTH_MATRIX dTildeWaveNoSum;

	bool shift;
	shift = inversionData->outputModel->dTilde.shift;
	dTildeWave = newMapDepthMatrix(input->K.nX, input->K.nY, 1, input->K.nT, shift);
	if (inversionData->debug) {
		dTildeWaveNoSum = newMapDepthMatrix(input->K.nX, input->K.nY, input->K.nZ, input->K.nT, shift);
	}

	MATRIX kernelk, qk;
	int ikx = input->K.nX;
	int iky = input->K.nY;
	int ii, jj; // indices for kx and ky respectively
	for (ii = 0; ii < ikx; ii++) {
		for (jj = 0; jj < iky; jj++) {
			kernelk = extractMapDepth(ii, jj, input->K);
			qk = extractMapDepth(ii, jj, inversionData->outputModel->bFull);
			MATRIX dTilda;
			// calculate d~
			// qk will always be a 1x(Nz x Nv) matrix; and we don't need to transpose it,
			// because we changed the number of maps after reading to be (Nz x Nv)

			dTilda = multiplyMM2('N', 'N', kernelk, qk);

			if (inversionData->debug) {
				MATRIX dTildaNoSum;
				MATRIX qkt = transpose(qk);
				dTildaNoSum =  starMultiply(kernelk, qkt);
				insertMapDepth(ii, jj, dTildaNoSum, &dTildeWaveNoSum);
				freeMatrix(dTildaNoSum);
				freeMatrix(qkt);
			}

			// calculation with Theta done at initialization

			// insert noise-free map into temporary dTilda matrix
			insertMapDepth(ii, jj, dTilda, &dTildeWave);
			freeMatrix(dTilda);
		}
	}

	// Inverse transform each dTilda
	MAP_DEPTH_MATRIX dTildeReal;
	dTildeReal = convertMapDepthSpace(&dTildeWave, REAL_SPACE);
	destroyMapDepthMatrix(dTildeWave);

	if (inversionData->debug) {
		MAP_DEPTH_MATRIX dTildeRealNoSum;
		dTildeRealNoSum = convertMapDepthSpace(&dTildeWaveNoSum, REAL_SPACE);
		FITSParameters headerParameters;
		headerParameters.n = 0;
		char **filenames;
		addParmsToFilename2(inversionData->outputBaseDir, NULL, 0, false);
		filenames = finishFilenamesWExt(inversionData->outputBaseDir, "fwdDepths", 1, false, false, Constants.extension);
		printf("writing to %s\n", filenames[0]);
		writeMapDepthMatrix(filenames, 1, dTildeRealNoSum, false, headerParameters);
		destroyMapDepthMatrix(dTildeRealNoSum);
		destroyMapDepthMatrix(dTildeWaveNoSum);
	}
	tempd->dTildeFwd = dTildeReal;
}

/**
 * Set random generator seed from input file if possible,
 * otherwise, use the system time.  Write the seed to an output file
 * @param[in] outputBaseDir The BaseDirectory for all output files.
 * @param[in] lambdaSet The random seed comes from the input parameter
 * file as the minimum value in a LambdaSet.
 */
void setSeed(BaseDirectory *outputBaseDir, const LambdaSet *lambdaSet) {

	unsigned int seed2;
	unsigned int seed;
    bool seedSet;

    seedSet = false;

	// try to read input seed
	if (lambdaSet->min < 0) {
		reportError(RandomSeedNegative, "setSeed", WARNING);
	} else if (lambdaSet->n != 0) {
		seed = (unsigned int) lambdaSet->min;
		seed2 =  setRandomSeed(seed);
		seedSet = true;
		if (seed2 != seed) {
			reportError(RandomSeedBad, "setSeed", WARNING);
		}
		seed = seed2;
	} else {
			reportError(RandomSeedSysTime, "setSeed", INFO);
	}
    if (!seedSet)
    	seed = setRandomSeedToTime();

	// write seed to output file; setup filename for file containing seed
	char **filenames;
	char *seedFilename;
	addParmsToFilename2(outputBaseDir, NULL, 0, false);
	filenames = finishFilenamesWExt(outputBaseDir, "seed", 1, false, false, Constants.txtExtension);
	seedFilename = filenames[0];
	FILE *fp;
	fp = fopen(seedFilename, "w");
	fprintf(fp, "%u\n", seed);
	fclose(fp);
}

/** Fill InversionData structure with data from input file.
 * Read kernels, model, and noise level parameter
 * @param[in] inputParms InputParms structure filled with filenames, directories and parameters
 * extracted from input file.
 * @param[in] commandLineOutputDir character string with new output directory specified on command line.
 * @param[in] debug	Boolean flag for writing intermediate matrix values.
 * @return InversionData structure filled with input data.
 */
InversionData *ForwardInit(InputParms inputParms, char *commandLineOutputDir,
		bool debug) {

	int dim;
	InversionData *inversionData;

	inversionData = (InversionData *) td_malloc(sizeof(InversionData));
	inversionData->debug = debug;

	// set problem dimension, which tells us how to interpret the
	// file map and parameters
	dim = getProblemDimension();
	if (commandLineOutputDir == NULL) {
		inversionData->outputBaseDir = newBaseDirectory(
				inputParms.outputBaseDirName);
	} else {
		inversionData->outputBaseDir = newBaseDirectory(commandLineOutputDir);
	}

	// read input files, K, A, q (model containing velocities); need read and write fitsio as a library.
	// 3d problem, read, transform and coalesce kernel files into giant matrix
	bool useFourierTransform = true;
	bool useFourierTransformA = false;
	printf("Reading Kernel files ...\n");
	MAP_DEPTH_MATRIX K = readMapDepthMatrixKernel(
			inputParms.d3.InputKernelBaseDirName,
			inputParms.d3.KernelFilenames, inputParms.d3.nKernelMapFiles,
			useFourierTransform);

	TdmMultiplyScalarTakeConjugate(&K, (2.0 * M_PI) * (2.0 * M_PI));

	// Read, transform and coalesce noise amplitude files into giant matrix?
	printf("Reading noise amplitude files ...\n");
	MAP_DEPTH_MATRIX A = readMapDepthMatrix(
			inputParms.d3.InputMapBaseDirName, inputParms.d3.MapFilenames,
			inputParms.d3.nKernelMapFiles, useFourierTransformA);
    //  add noise after maps have been calculated and inverse transformed.

	printf("Reading model file ...\n");
	MAP_DEPTH_MATRIX b = readMapDepthMatrixFromOneFile(inputParms.InputICovarFileName, useFourierTransform);
	//                                    normal ordering:      (Nx,Ny,Nz,M,Nv)
	// swap indices because readMapDepthMatrixFromOneFile fills (Nx,Ny,Nz,Nv,1)
	// but we want                                              (Nx,Ny,Nz,1,Nv)
	// so swap M and Nv
	int save = b.nT;
	b.nT = b.nV;
	b.nV = save;

	printf("Reading theta file ...\n");
	// read zgrid file and construct Theta
	// handle Theta for 3D and 2D problems.  If 2D, set Theta to 1
	MATRIX Theta;
	if (dim > PDC.D2 ) {
		// Read Theta; overlap matrix
		if (inputParms.InputRegFileName != NULL)
			Theta = toMatrix(readfits(inputParms.InputRegFileName));
	} else {
		Theta = newMatrix(1,1);
		double *p;
		p = Theta.data;
		p[0] = 1.0;
	}

	// check matrix dimensions for compatibility
	printf("checking dimensions ...\n");
	if (!DimensionsOkFwd(K, b, A, Theta, useFourierTransform))
		reportError(IncompatibleMatrices, "ForwardInit", FATAL);
    printf("  dimensions OK\n");

	// set seed for random number generator
	setSeed(inversionData->outputBaseDir, &inputParms.lambdaSet2);

	freeInputParms(inputParms);

	Model *model;
	model = (Model *) td_malloc(sizeof(Model));
	inversionData->outputModel = model;
	// initialize d~ matrix and b matrix
	initializeForwardOutputModel(model, K);

	// store the model
	/* HACK to avoid a transpose of qk */
	b.nT = b.nV * b.nZ;
	b.nV = 1;
	b.nZ = 1;
	/* end HACK */
	inversionData->outputModel->bFull = b;

	// convert Theta to complex for homogeneous matrix multiplication
	MATRIX temp = ConvertToComplex(Theta);
	freeMatrix(Theta);
	Theta = temp;

	// apply overlap (basis vectors) to Kernel before further processing
	MdmMultiplyMatrix(&K, Theta);
	freeMatrix(Theta);

	inversionData->InputMatrices = (InputData *) td_malloc(sizeof(InputData));
	InputData *id;
	id = inversionData->InputMatrices;
	inversionData->IntermediateMatrices = (TempData *) td_malloc(
			sizeof(TempData));

	id->K = K;
	id->M = A;  // put the noise amplitudes in place of the maps

	LambdaSet muSet = inputParms.lambdaSet1;
	LambdaSet lambdaSet2;
	lambdaSet2.min = 0.0;
	lambdaSet2.max = 1.0;
	lambdaSet2.n = 1;

	inversionData->parameter1 = lambdaSet2;
	inversionData->parameter2 = muSet;
	inversionData->nthStepParm1 = 1;
	inversionData->nthStepParm2 = 1;
	inversionData->currentValueParm1 = lambdaSet2.min;
	inversionData->currentValueParm2 = muSet.min;

	// Now perform forward calculation
	ForwardCalc(inversionData);

	return (inversionData);
}


/**
 * Add noise to maps and stuff them into dTilde in the output model.
 * Setup the output files with the parameter steps. Record
 * parameter values in model and inversion data structure.
 * @param[in,out] inversionData Pointer to InversionData structure, which
 * contains all input data, temporary data, parameters and result data.
 */
void ForwardAddNoise(InversionData *inversionData) {
	// loop around add noise, because dTilda doesn't change

	InputData *input;
	input = inversionData->InputMatrices;
	TempData *tempd;
	tempd = inversionData->IntermediateMatrices;


	double mu;
	mu = inversionData->currentValueParm2;
	setParameters2(inversionData->outputModel, &mu, 1);
	bool debugPosition = false;
	int values[1];
	values[0] = inversionData->nthStepParm2;
	addParmsToFilename2(inversionData->outputBaseDir, values, 1, debugPosition);

    MAP_DEPTH_MATRIX dTilda;

    dTilda = tempd->dTildeFwd;
    MAP_DEPTH_MATRIX A;
    A = input->M;

	int Nx;
	Nx = dTilda.nX;
	int M;
	M = dTilda.nT;

	MATRIX r;

	printf("noise scale = %lf\n", mu);

	MAP_DEPTH_MATRIX A2;
	if (mu > 0.0) {
		// copy A and multiply it by mu
		A2 = TdmCopy(A);
		TdmMultiplyScalar(&A2, mu);
	}

	// add noise: d~ = d~ + (mu * A) .* randn
	int m;
	for (m=0; m<M; m++) {
	  // extract d~
	  MATRIX d;
	  d = extractXYPlaneFromMapDepthMatrix(m, 0, dTilda);
#ifdef TESTING
		// get random matrix
		r = newRandomMatrix(Nx,Nx);
#endif
      if (mu > 0.0) {
#ifndef TESTING
		// get random matrix
		r = newRandomMatrix(Nx,Nx);
#endif
		// multiply individual components
		MATRIX Axyr;
		MATRIX Axy;
		// need to extract A(x,y) from A
		Axy = extractXYPlaneFromMapDepthMatrix(m, 0, A2);
		Axyr = starMultiply(Axy, r);

		MATRIX dn = newComplexMatrix(d.nRow, d.nCol);
		// add noise to  d~
		APlusLambdaB(d, 1.0, Axyr, dn);
		// insert into new d~ matrix
		insertXYPlaneIntoMapDepthMatrix(dn, m, 0, &(inversionData->outputModel->dTilde));
#ifndef TESTING
		freeMatrix(r);
#endif
		freeMatrix(Axy);
		freeMatrix(dn);
		freeMatrix(Axyr);
      } else {
		insertXYPlaneIntoMapDepthMatrix(d, m, 0, &(inversionData->outputModel->dTilde)); // Just for testing
	  }
	  freeMatrix(d);
#ifdef TESTING
		freeMatrix(r);
#endif
	}

	if (mu > 0.0)
 	  destroyMapDepthMatrix(A2);
}

/**
 * Complete forward calculation by writing d~ matrix.
 * @param[in] inversionData Pointer to data:
 *  K 4D-matrix containing Kernel data in wavespace.
 *  b 4D-matrix containing model data, b, in wavespace.
 *  d~ 4D-matrix containing estimate of travel-time map, in wavespace.
 *  baseDir BaseDirectory structure which manages the output filenames.
 *  debug Boolean flag: true if writing intermediate matrix values during inversion.
 */
void ForwardComplete(InversionData *inversionData) {
	Model *model;
	BaseDirectory *baseDir;

	model = inversionData->outputModel;
	baseDir = inversionData->outputBaseDir;

	// write model.write d~
	bool useFT = false;
	writeDTildaValues3d(model, baseDir, useFT);
	printf("finished writing set of maps\n");
}

/**
 * Free memory after forward calculation is complete.
 * @param[in,out] inversionData Pointer to InversionData structure.
 */
void ForwardDestroy(InversionData *inversionData) {

	InputData *im;
	im = inversionData->InputMatrices;
	destroyMapDepthMatrix(im->K);
	destroyMapDepthMatrix(im->M);

	Model *model = inversionData->outputModel;

	destroyMapDepthMatrix(model->bFull);
	destroyMapDepthMatrix(model->dTilde);
	td_free(model, sizeof(Model));

	TempData *tempd;
	tempd = inversionData->IntermediateMatrices;
	destroyMapDepthMatrix(tempd->dTildeFwd);

	destroyBaseDirectory(inversionData->outputBaseDir);

	td_free(inversionData->InputMatrices, sizeof(InputData));
	td_free(tempd, sizeof(TempData));
	td_free(inversionData, sizeof(InversionData));

	printf("Memory Usage  %ld \tFinal\n", getMemoryReport());

	exit(EXIT_SUCCESS);

}

bool ForwardMoreParameterSteps(InversionData *inversionData) {
	return ((inversionData->nthStepParm1 <= inversionData->parameter1.n) &&
	(inversionData->nthStepParm2 <= inversionData->parameter2.n));
}

void ForwardNextParameter(InversionData *inversionData) {

	LambdaSet muRange;

	muRange = inversionData->parameter2;

	inversionData->nthStepParm2 += 1;
	if (inversionData->nthStepParm2 > inversionData->parameter2.n) {
		inversionData->nthStepParm1 += 1;
		inversionData->currentValueParm1 =
				nextLambdaValue(inversionData->parameter1, inversionData->nthStepParm1-1);
		inversionData->nthStepParm2 = 1;
		inversionData->currentValueParm2 = muRange.min;

	} else {
		inversionData->currentValueParm2 =
				nextLambdaValue(inversionData->parameter2, inversionData->nthStepParm2-1);
	}
}

